"""Pycatzao is a pure Python library for encoding, decoding and compressing Asterix CAT240 messages.

Below, find the documentation of the methods exposed by the API of Pycatzao grouped
under the topics "Decoder", exposed as:

 - `pycatzao.decode`
 - `pycatzao.decode_file`

for decoding already existing bytestreams of Asterix CAT240 messsages, "Encoder",
exposed as

 - `pycatzao.encode`
 - `pycatzao.make_summary`
 - `pycatzao.make_video_header`
 - `pycatzao.make_video_message`

for compiling new messages, and "Compressor", exposed as:

 - `pycatzao.compress`
 - `pycatzao.compress_file`

for compressing (encoded) messages.

Typically, no interactions between methods of different topics is necessary and one can
safely study their documentation separately.

Convenient helper functions for inferring the implicit binning scheme or for stacking
decoded block are exposed as "Utilities":

 - `pycatzao.infer_bin_edges`
 - `pycatzao.join_blocks`

Example:
    .. code-block:: python

        buffer_size = 100_000
        join_n_blocks = 1_000_000

        # compress data
        data_file = "my-data.cat240"
        with open("compressed.cat240", "wb") as f:
            for msg in pycatzao.compress_file(data_file, buffer_size=buffer_size):
                f.write(msg)

        # convert to csv
        for i, blocks in enumerate(
            itertools.batched(
                pycatzao.decode_file("compressed.cat240", buffer_size=buffer_size),
                join_n_blocks,
            )
        ):
            pd.DataFrame(pycatzao.join_blocks(blocks)).to_csv(
                f"csv-{i + 1}.csv.gz", header=True, index=False, float_format="%.2f"
            )

"""  # noqa: E501

from . import _version

__version__ = _version.get_versions()["version"]

# export public API of the package
from .compress import (
    compress,  # noqa: F401
    compress_file,  # noqa: F401
)
from .decoder import (
    decode,  # noqa: F401
    decode_file,  # noqa: F401
)
from .encoder import (
    encode,  # noqa: F401
    make_summary,  # noqa: F401
    make_video_header,  # noqa: F401
    make_video_message,  # noqa: F401
)
from .utils import (
    infer_bin_edges,  # noqa: F401
    join_blocks,  # noqa: F401
)



#!/usr/bin/env python

import os
import sys

from setuptools import find_packages, setup

here = os.path.abspath(os.path.dirname(__file__))
sys.path.append(here)

import versioneer  # noqa: E402

with open(os.path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()


setup(
    name="pycatzao",
    description="A pure Python library for encoding, decoding and compressing Asterix"
    " CAT240 messages.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/DLR-KN/pycatzao",
    author="Nis Meinert",
    author_email="nis.meinert@dlr.de",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    packages=find_packages(exclude=["docs", "tests"]),
    python_requires=">=3.10",
    install_requires=[
        "numpy",
        "tqdm",
    ],
    extras_require={
        "dev": [
            "pre-commit",
            "pytest",
            "pytest-cov",
            "sphinx",
        ],
    },
)



"""Decoding of Asterix CAT240 messages.

Use :func:`decode` for decoding a byte sequence of encoded Asterix CAT240 messages. This
function will return trailing bytes that were not decoded such that one can easily
implement loops without the need to split CAT240 messages beforehand. Below, we show a
simple example how to decode data from a file in chunks. (Note that this is just a dull
example to demonstrate how to utilize the returned state, checkout :func:`decode_file`
if you actually want to decode a file!)

Example:
    .. code-block:: python

        with open("my-cat240-data.bin", "rb") as f:
            state = b""
            while chunk := f.read(1024)) != b"":
                blocks, state = pycatzao.decode(state + chunk)
                for block in blocks:
                    process_blocks(block)  # do somehting meaningful with the data

Here, a decoded `block` is a `dict` that contains the payload of the respective CAT240
message, i.e.,

 - `summary`: ASCII string to deliver stream meta data and/or labels

for blocks of message type `001` (Video Summary message), and

 - `idx`: Message sequence identifier
 - `az`: The azimuth in degrees. This value is calculated as the circular mean of `START_AZ` and `END_AZ` as both values are often set to non-meaningful values and it is very easy to confuse their range with the sensor resolution. In contrast, their circular mean is less ambiguous and can be used safely for downstream tasks.
 - `r`: Range of the data points in meters
 - `amp`: Video signal amplitude (aka the "data points")

for message type `002` (Video message), as well as

 - `sac`: System Area Code (SAC)
 - `sic`: System Identification Code (SIC)
 - `type`: Type of the message, `1` (Video summary message) or `2` (Video message)
 - `tod` (if present): Time of Day in seconds (We literally just decode the encoded value. It is beyond the scope of this library to interpret this value, e.g., as an absolute UTC time stamping.)

in both messages.
"""  # noqa: E501

import zlib

import numpy as np

from pycatzao import _utils


def _decode_block(data):
    uap1 = data[3]
    m040 = uap1 & 0x08 == 0x08
    mFX = uap1 & 0x01 == 0x01

    i, uap2 = (5, data[4]) if mFX else (4, 0)
    m050 = uap2 & 0x40 == 0x40
    m051 = uap2 & 0x20 == 0x20
    m052 = uap2 & 0x10 == 0x10
    m140 = uap2 & 0x08 == 0x08

    msg = {
        "sac": data[i],  # I240/010
        "sic": data[i + 1],  # I240/010
        "type": data[i + 2],  # I240/000
    }
    i += 3

    match msg["type"]:
        case 1:
            # I240/030
            n = data[i]
            msg["summary"] = data[i + 1 : i + 1 + n].decode("ascii")
            i += 1 + n

        case 2:
            # I240/020
            msg["idx"] = int.from_bytes(data[i : i + 4], byteorder="big")
            i += 4

            # I240/040 or I240/041
            az_scale = 360 / 2**16
            t_scale = 1e-9 if m040 else 1e-15

            start_az = int.from_bytes(data[i : i + 2], byteorder="big") * az_scale
            end_az = int.from_bytes(data[i + 2 : i + 4], byteorder="big") * az_scale
            start_rg = int.from_bytes(data[i + 4 : i + 8], byteorder="big")
            cell_dur = int.from_bytes(data[i + 8 : i + 12], byteorder="big") * t_scale
            i += 12

            msg["az"] = _utils._circular_mean(start_az, end_az)

            # I240/048
            compression = data[i] & 0x80 == 0x80
            res = 2 ** (data[i + 1] - 1)
            i += 2

            # I240/049
            nb_vb = int.from_bytes(data[i : i + 2], byteorder="big")
            nb_cells = int.from_bytes(data[i + 2 : i + 5], byteorder="big")
            i += 5

            # I240/050 or I240/051 or I240/052
            if m050:
                n = 4
            elif m051:
                n = 64
            elif m052:
                n = 256
            else:  # pragma: no cover
                raise ValueError("Invalid Asterix CAT240 message.")

            n *= data[i]
            amp = data[i + 1 : i + n + 1][:nb_vb]
            i += n + 1

            if compression:
                try:
                    amp = zlib.decompress(amp, wbits=0)
                except zlib.error as e:  # pragma: no cover
                    raise ValueError("Video blocks not compressed with zlib.") from e

            if res == 8:
                dtype = np.uint8
            elif res == 16:
                dtype = np.uint16
            elif res == 32:
                dtype = np.uint32
            else:  # pragma: no cover
                raise NotImplementedError(
                    f"{res} bit resolution is not yet implemented."
                )

            msg["amp"] = np.frombuffer(amp, dtype=dtype)[:nb_cells]

            c = 299_792_458
            msg["r"] = cell_dur * (start_rg + np.arange(nb_cells)) * c / 2

            mask = msg["amp"] > 0
            msg["amp"] = msg["amp"][mask]
            msg["r"] = msg["r"][mask].astype(np.float32)

    # I240/140
    if m140:
        msg["tod"] = int.from_bytes(data[i : i + 3], byteorder="big") / 128
        i += 3

    if i != len(data):  # pragma: no cover
        raise ValueError("Invalid Asterix CAT240 message.")

    return msg


def decode(data):
    r"""Decode CAT240 data.

    This functions decodes a given binary blob of encoded Asterix CAT240 messages (type
    `001` or `002`). The data have to start with a new CAT240 message but can end in
    the middle of one. The bytes of the incomplete message at the end of the input data
    are returned such that a chunking a long byte sequence by subsequent calls to
    :func:`decode` becomes trivial.

    Example:
        >>> data = b'\\xf0\\x00\\x13\\xd1...'
        >>> blocks, state = pycatzao.decode(data[:128])
        >>> blocks
        [{'sac': 7, 'sic': 42, 'type': 1, 'summary': '...', 'tod': 100.0},
         {'sac': 7, 'sic': 42, 'type': 2, 'idx': 4711, 'az': 5.9, 'amp': array([248, ...], dtype=uint8), 'r': array([ 94.2, ...], dtype=float32), 'tod': 100.0},
         {'sac': 7, 'sic': 42, 'type': 2, 'idx': 4712, 'az': 6.4, 'amp': array([127, ...], dtype=uint8), 'r': array([ 89.2, ...], dtype=float32), 'tod': 100.1}]
        >>> blocks, state = pycatzao.decode(state + data[128:])
        >>> blocks
        [{'sac': 7, 'sic': 42, 'type': 2, 'idx': 4713, 'az': 6.9, 'amp': array([255, ...], dtype=uint8), 'r': array([ 94.2, ...], dtype=float32), 'tod': 100.2}, ...]

    Args:
        data (bytes):
            Encoded CAT240 messages as a binary blob that start with a new message.

    Returns:
        tuple[dict, bytes]:
            Decoded message and trailing bytes. The latter carries the state of the
            decoder and should be prepended to the input data for the subsequent call
            to :func:`decode`.
    """  # noqa: E501
    return _utils._map_blocks(data, func=_decode_block)


def decode_file(file_name, *, size=-1, buffer_size=-1):
    """Decode a CAT240 file.

    This is a helper function that reads and decodes a file with (binary) Asterix
    CAT240 data.

    Args:
        file_name (str | pathlib.Path):
            Name of the file.
        size (int):
            Maximum number of bytes to read from the file. If negative, the entire file
            will be read.
        buffer_size (int):
            Process file in chunks of this size. If negative, the entire file will be
            loaded to RAM at once.

    Returns:
        typing.Generator[dict, None, None]:
            Decoded messages (type of generated items is the same as the first return
            type of :func:`decode`.)
    """
    yield from _utils._map_file(
        file_name, func=decode, size=size, buffer_size=buffer_size
    )



r"""Encoding of Asterix CAT240 messages.

Use :func:`encode` to compile new Asterix CAT240 messages and feed it with the output of
either :func:`make_summary` or :func:`make_video_message` for creating a type `001` or
`002` message, respectively. The output is a byte sequence of a single CAT240 message
that can directly be concatenated with others:

Example:
    .. code-block:: python

        payloads = [
            pycatzao.make_summary(summary="FoObaR"),
            pycatzao.make_video_message(
                np.array([255, 0, 255], dtype=np.uint8),  # signal at range = [9m, 12m, 15m]
                msg_index=1,
                header=pycatzao.make_video_header(
                    start_az=45.0,
                    end_az=45.0,
                    cell_offset=9,  # (center of) first cell is at 9m
                    cell_width=3,   # cell width is 3m
                )
            ),
        ]

        tod = ...  # get ToD in seoncds
        blocks = [pycatzao.encode(msg, sac=7, sic=42, tod=tod) for msg in payloads]
        print(b"".join(blocks).hex())  # outputs: f00012d108072a01...
"""  # noqa: E501

import zlib

import numpy as np


def make_summary(summary):
    """Create Video Summary.

    Creates a type `001` message (Video Summary message) to be passed as the payload to
    :func:`encode`.

    Args:
        summary (str):
            (ASCII) string to deliver stream meta data and/or labels.

    Returns:
        typing.Any:
            Payload for :func:`encode`.

    Raises:
        ValueError:
            The ASCII representation of the summary has to be not longer than 0xFFFF
            characters.
    """
    chars = summary.encode("ascii")
    n = len(chars)
    if n > 0xFFFF:  # pragma: no cover
        raise ValueError(
            f"Summary message is too long. Not more than {0xFFFF} ASCII symbols are"
            " allowed. See specification of I240/030 for details."
        )

    return {
        "uap": 0x1000,
        "data": n.to_bytes(1, byteorder="big") + chars,
    }


def make_video_header(*, start_az, end_az, cell_offset, cell_width):
    """Create Video Header.

    Creates a video header (I240/040 or I240/041) for :func:`make_video_message`.

    Args:
        start_az (float):
            Start azimuth in degrees.

        end_az (float):
            End azimuth in degrees.

        cell_offset (float):
            Offset of first cell in meters.

        cell_width (float):
            Cell width in meters. If this value is larger than 500m, the header type
            is I240/040 (Video Header Nano) and I240/041 (Video Header Femto) otherwise.

    Returns:
        typing.Any:
            The header for func:`make_video_message`.
    """
    start_az %= 360
    end_az %= 360

    az_scale = 2**16 / 360
    start_az = round(start_az * az_scale).to_bytes(2, byteorder="big")
    end_az = round(end_az * az_scale).to_bytes(2, byteorder="big")

    item, t_scale = ("I240/040", 1e-9) if cell_width > 500 else ("I240/041", 1e-15)
    cell_dur = round(2 * cell_width / (299_792_458 * t_scale)).to_bytes(
        4, byteorder="big"
    )
    start_rg = round(cell_offset / cell_width).to_bytes(4, byteorder="big")

    return {
        "uap": 0x0800 if item == "I240/040" else 0x0400,
        "data": start_az + end_az + start_rg + cell_dur,
    }


def make_video_message(amp, *, msg_index, header, compress=True):
    """Create Video Message.

    Creates a video message (I240/050, I240/051 or I240/052) for :func:`encode` from a
    video header.

    Args:
        amp (np.ndarray):
            Amplitude array given as an `np.ndarray` with `dtype` set to either
            `np.uint8`, `np.uint16`, or `np.uint32`.

        msg_index (int):
            Message Sequence Identifier (video record cyclic counter).

        header:
            Header generated by :func:`make_video_header`.

        compress (bool):
            Compress the amplitude array before compiling the message. Note that this is
            typically cheap yet very effective and should be enabled unless you know
            better.

    Returns:
        typing.Any:
            Payload for :func:`encode`.
    """
    if not isinstance(amp, np.ndarray):  # pragma: no cover
        raise ValueError("Amplitude array has to be of type `np.ndarray`.")

    if amp.dtype not in [np.uint8, np.uint16, np.uint32]:  # pragma: no cover
        raise ValueError(
            "Data type of amplitude array is not supported. Data type has to be either"
            " 'np.uint8', 'np.uint16', or 'np.uint32'."
        )

    nb_cells = len(amp)
    amp = amp.tobytes()

    res = ((len(amp) * 8) // nb_cells).bit_length().to_bytes(1, byteorder="big")

    if compress:
        amp = zlib.compress(amp)

    nb_vb = len(amp)

    items = {4: "I240/050", 64: "I240/051", 256: "I240/052"}
    keys = list(items.keys())
    i = np.argmin([(1 + (nb_vb - 1) // n) * (n + 1) for n in keys])
    n = keys[i]
    item = items[n]

    amp += b"\x00" * ((1 + (nb_vb - 1) // n) * n - nb_vb)
    assert len(amp) % n == 0

    n = (len(amp) // n).to_bytes(1, byteorder="big")
    nb_vb = nb_vb.to_bytes(2, byteorder="big")
    nb_cells = nb_cells.to_bytes(3, byteorder="big")

    uap = {"I240/050": 0x02C0, "I240/051": 0x02A0, "I240/052": 0x0290}[item]
    uap = uap | header["uap"]

    data = msg_index.to_bytes(4, byteorder="big")
    data += header["data"]
    data += (b"\x80" if compress else b"\x00") + res + nb_vb + nb_cells + n + amp

    return {
        "uap": uap,
        "data": data,
    }


def encode(payload, *, sac, sic, tod=-1):
    """Compiles data into a CAT240 message.

    Takes a payload generated by either :func:`make_summary` or
    :func:`make_video_message` and wraps it as an Asterix CAT240 message.

    Args:
        payload: The payload to encode.
        sac (int):
            System Area Code (SAC)
        sic (int):
            System Identification Code (SIC)
        tod (float):
            Time of Day (ToD) in seconds. If negative, no ToD is included into the
            compiled message.

    Returns:
        bytes:
            A single Asterix CAT240 message.
    """
    sac = sac.to_bytes(1, byteorder="big")
    sic = sic.to_bytes(1, byteorder="big")
    msg_type = b"\x01" if payload["uap"] & 0x1000 else b"\x02"

    if tod <= 0:
        tod = b""
    else:
        tod = round(tod * 128)
        if tod > 0xFFFFFF:  # pragma: no cover
            raise ValueError(
                "Time of Day (given in seconds) is too large. See specification of"
                " I240/140 for details."
            )

        tod = tod.to_bytes(3, byteorder="big")

    uap = 0xC100 if tod == b"" else 0xC108
    uap |= payload["uap"]
    uap = uap.to_bytes(2, byteorder="big")

    data = uap + sac + sic + msg_type + payload["data"] + tod

    return b"\xf0" + (len(data) + 3).to_bytes(2, byteorder="big") + data
